==> reytool-0.65-py3-none-any.whl: reytool/rdatabase.py <==
# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
@Time    : 2022-12-05 14:10:02
@Author  : Rey
@Contact : reyxbo@163.com
@Explain : Rey's database connect type
"""


from typing import Any, List, Dict, Iterable, Optional, Literal, Union
import re
from sqlalchemy import create_engine, text
from sqlalchemy.engine.base import Engine, Connection
from sqlalchemy.engine.cursor import LegacyCursorResult
from sqlalchemy.sql.elements import TextClause

from .rbasic import get_first_notnull, error
from .rdata import to_table, to_df, to_json, to_sql, to_html, to_csv, to_excel
from .rtext import rprint
from .rwrap import runtime


def monkey_patch_more_fetch() -> None:
    """
    Add more methods to LegacyCursorResult object of sqlalchemy package.
    """

    # Fetch SQL result to table in List[Dict] format.
    LegacyCursorResult.fetch_table = to_table

    # Fetch SQL result to DataFrame object.
    LegacyCursorResult.fetch_df = to_df

    # Fetch SQL result to JSON string.
    LegacyCursorResult.fetch_json = to_json

    # Fetch SQL result to SQL string.
    LegacyCursorResult.fetch_sql = to_sql

    # Fetch SQL result to HTML string.
    LegacyCursorResult.fetch_sql = to_html

    # Fetch SQL result to save csv format file.
    LegacyCursorResult.fetch_csv = to_csv

    # Fetch SQL result to save excel file.
    LegacyCursorResult.fetch_excel = to_excel

monkey_patch_more_fetch()

class RConn(object):
    """
    Rey's database connect type.
    """

    # Values to be converted to None.
    none_values: List = ["", " ", b"", [], (), {}, set()]
    
    def __init__(
        self,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[str] = None,
        database: Optional[str] = None,
        charset: Optional[str] = None,
        once: bool = True,
        conn: Optional[Union[Engine, Connection]] = None
    ) -> None:
        """
        Set database connect parameters.
        """

        if type(conn) == Connection:
            conn = conn.engine
        if type(conn) == Engine:
            user = get_first_notnull(user, conn.url.username)
            password = get_first_notnull(password, conn.url.password)
            host = get_first_notnull(host, conn.url.host)
            port = get_first_notnull(port, conn.url.port)
            database = get_first_notnull(database, conn.url.database)
            charset = get_first_notnull(charset, conn.url.query.get("charset"))
            conn = conn.connect(close_with_result=once)
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.database = database
        self.charset = charset
        self.once = once
        self.conn = conn

    def create_conn(
        self,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[Union[str, int]] = None,
        database: Optional[str] = None,
        charset: Optional[str] = None,
        once: bool = None
    ) -> Connection:
        """
        Get database connect engine.
        """

        if self.conn != None\
            and (user == None or self.conn.engine.url.username == user) \
            and (password == None or self.conn.engine.url.password == password) \
            and (host == None or self.conn.engine.url.host == host) \
            and (port == None or self.conn.engine.url.port == port) \
            and (database == None or self.conn.engine.url.database == database) \
            and (charset == None or self.conn.engine.url.query["charset"] == charset):
            return self.conn

        user: str = get_first_notnull(user, self.user, default="error")
        password: str = get_first_notnull(password, self.password, default="error")
        host: str = get_first_notnull(host, self.host, default="error")
        port: Union[str, int] = get_first_notnull(port, self.port, default="error")
        database: str = get_first_notnull(database, self.database, default="error")
        charset: str = get_first_notnull(charset, self.charset, default="utf8")
        once: str = get_first_notnull(once, self.once, default=True)

        try:
            url = f"mysql+mysqldb://{user}:{password}@{host}:{port}/{database}?charset={charset}"
            engine = create_engine(url)
        except ModuleNotFoundError:
            url = f"mysql+pymysql://{user}:{password}@{host}:{port}/{database}?charset={charset}"
            engine = create_engine(url)
        conn = engine.connect(close_with_result=once)
        if self.conn == None and not once:
            self.conn = conn
        return conn
    
    def fill_sql_none(
        self,
        sql: Union[str, TextClause],
        parms: Union[Dict, List[Dict]],
        fill_field: bool = True,
        none_values: List = none_values
    ) -> Union[Dict, List[Dict]]:
        """
        Fill missing parameters according to contents of sqlClause object of sqlalchemy module.

        Parameters
        ----------
        sql : SQL in sqlalchemy.text format or return of sqlalchemy.text.
        parms : Parameters set for filling sqlalchemy.text.
        fill_field : Whether fill missing fields.
        none_values : Values to be converted to None.

        Returns
        -------
        Filled parameters.
            - When parameter parms is Dict, then return Dict.
            - When parameter parms is List[Dict], then return List[Dict].
        """

        if type(sql) == TextClause:
            sql = sql.text
        pattern = "(?<!\\\):(\w+)"
        sql_keys = re.findall(pattern, sql)
        if type(parms) == dict:
            for key in sql_keys:
                if fill_field:
                    val = parms.get(key)
                else:
                    val = parms[key]
                if val in none_values:
                    val = None
                parms[key] = val
        else:
            for parm in parms:
                for key in sql_keys:
                    if fill_field:
                        val = parm.get(key)
                    else:
                        val = parm[key]
                    if val in none_values:
                        val = None
                    parm[key] = val
        return parms

    def execute(
        self,
        sql: str,
        parms: Optional[Union[List[Dict], Dict]] = None,
        database: Optional[str] = None,
        fill_field: bool = True,
        none_values: List = none_values,
        once: bool = None,
        report: bool = False,
        **kw_parms: Any
    ) -> LegacyCursorResult:
        """
        Execute SQL.

        Parameters
        ----------
        sql : SQL in sqlalchemy.text format.
        parms : Parameters set for filling sqlalchemy.text.
        database : Database name.
        fill_field : Whether fill missing fields.
        none_values : Values to be converted to None.
        once : Whether the database connect engine is one-time (i.e. real time creation).
        report : Whether print SQL and SQL runtime.
        kw_parms : Keyword parameters for filling sqlalchemy.text.

        Returns
        -------
        LegacyCursorResult object of alsqlchemy package.
        """

        once = get_first_notnull(once, self.once, default=True)

        conn = self.create_conn(database=database, once=once)
        if parms != None:
            if type(parms) != list:
                parms = [parms]
            else:
                parms = parms.copy()
            for parm in parms:
                parm.update(kw_parms)
        else:
            parms = kw_parms
        if parms != {}:
            parms = self.fill_sql_none(sql, parms, fill_field, none_values)
        sql = text(sql)
        if report:
            result, report_runtime = runtime(conn.execute, sql, parms, _ret_report=True)
            rowcount = result.rowcount
            report = "%s\nRow Count: %d\n\n%s\n\n%s" % (report_runtime, rowcount, sql, parms)
            rprint(report, title="SQL", frame="half")
        else:
            result = conn.execute(sql, parms)
        if once:
            conn.close()
            del self.conn
            self.conn = None
        return result

    def execute_select(
            self,
            table: str,
            database: Optional[str] = None,
            fields: Optional[Union[str, Iterable]] = None,
            where: Optional[str] = None,
            order: Optional[str] = None,
            limit: Optional[Union[int, str, Iterable[Union[int, str]]]] = None,
            report: bool = False
        ) -> LegacyCursorResult:
        """
        Execute select SQL.

        Parameters
        ----------
        table : Table name.
        database : Database name.
        fields : Select syntax content.
            - None : Is 'SELECT *'.
            - str : Join as 'SELECT str'.
            - Iterable[str] : Join as 'SELECT \`str\`, ...'.

        where : 'WHERE' syntax content, join as 'WHERE str'.
        order : 'ORDER BY' syntax content, join as 'ORDER BY str'.
        limit : 'LIMIT' syntax content.
            - Union[int, str] : Join as 'LIMIT int/str'.
            - Iterable[Union[str, int]] with length of 1 or 2 : Join as 'LIMIT int/str, ...'.

        report : Whether print SQL and SQL runtime.

        Returns
        -------
        LegacyCursorResult object of alsqlchemy package.
        """

        sqls = []
        if database == None:
            _database = self.database
        else:
            _database = database
        if fields == None:
            fields = "*"
        elif type(fields) != str:
            fields = ",".join(["`%s`" % field for field in fields])
        select_sql = (
            f"SELECT {fields}\n"
            f"FROM `{_database}`.`{table}`"
        )
        sqls.append(select_sql)
        if where != None:
            where_sql = "WHERE %s" % where
            sqls.append(where_sql)
        if order != None:
            order_sql = "ORDER BY %s" % order
            sqls.append(order_sql)
        if limit != None:
            list_type = type(limit)
            if list_type in [str, int]:
                limit_sql = f"LIMIT {limit}"
            else:
                if len(limit) in [1, 2]:
                    limit_content = ",".join([str(val) for val in limit])
                    limit_sql = "LIMIT %s" % limit_content
                else:
                    error("The length of the limit parameter value must be 1 or 2", ValueError)
            sqls.append(limit_sql)
        sql = "\n".join(sqls)
        result = self.execute(sql, database=database, report=report)
        return result

    def execute_update(
        self,
        data: Union[LegacyCursorResult, List[Dict], Dict],
        table: str,
        database: Optional[str] = None,
        primary_key: Optional[str] = None,
        report: bool = False
    ) -> LegacyCursorResult:
        """
        Update the data of table in the datebase.
    
        Parameters
        ----------
        data : Updated data.
        table : Table name.
        database : Database name.
        primary_key : 'WHERE' syntax content.
            - None : Is 'WHERE `%s` = :%s' % (first field name and value of each item).
            - str : Is f'WHERE `{str}` = :{str}'.

        report : Whether print SQL and SQL runtime.

        Returns
        -------
        LegacyCursorResult object of alsqlchemy package.
        """

        data_type = type(data)
        if data_type == LegacyCursorResult:
            data = to_table(data)
        elif data_type == dict:
            data = [data]
        if database == None:
            _database = self.database
        else:
            _database = database
        data_flatten = {}
        sqls = []
        for index, row in enumerate(data):
            for key, val in row.items():
                index_key = "%d_%s" % (index, key)
                data_flatten[index_key] = val
            if primary_key == None:
                primary_key = list(row.keys())[0]
            set_content = ",".join(
                [
                    "`%s` = :%d_%s" % (key, index, key)
                    for key in row
                    if key != primary_key
                ]
            )
            sql = (
                f"UPDATE `{_database}`.`{table}`\n"
                f"SET {set_content}\n"
                f"WHERE `{primary_key}` = :{index}_{primary_key}"
            )
            sqls.append(sql)
        sqls = ";\n".join(sqls)
        result = self.execute(sqls, data_flatten, database, once=False, report=report)
        if self.once:
            self.conn == None
        return result

    def execute_insert(
        self,
        data: Union[LegacyCursorResult, List[Dict], Dict],
        table: str,
        database: Optional[str] = None,
        duplicate_method: Optional[Literal["ignore", "update"]] = None,
        report: bool = False
    ) -> LegacyCursorResult:
        """
        Insert the data of table in the datebase.

        Parameters
        ----------
        data : Updated data.
        table : Table name.
        database : Database name.
        duplicate_method : Syntax method when constraint error.
            - None : Then no syntax.
            - 'ignore' : Is 'UPDATE IGNORE INTO'.
            - 'update' : Is 'ON DUPLICATE KEY UPDATE'.

        report : Whether print SQL and SQL runtime.

        Returns
        -------
        LegacyCursorResult object of alsqlchemy package.
        """

        data_type = type(data)
        if data_type == LegacyCursorResult:
            data = self.to_table(data)
        elif data_type == dict:
            data = [data]
        if database == None:
            _database = self.database
        else:
            _database = database
        fields = list({key for row in data for key in row})
        fields_str = ",".join(["`%s`" % field for field in fields])
        fields_str_position = ",".join([":" + field for field in fields])
        if duplicate_method == "ignore":
            sql = (
                f"INSERT IGNORE INTO `{_database}`.`{table}`({fields_str})\n"
                f"VALUES({fields_str_position})"
            )
        elif duplicate_method == "update":
            update_content = ",".join(["`%s` = VALUES(`%s`)" % (field, field) for field in fields])
            sql = (
                f"INSERT INTO `{_database}`.`{table}`({fields_str})\n"
                f"VALUES({fields_str_position})\n"
                "ON DUPLICATE KEY UPDATE\n"
                f"{update_content}"
            )
        else:
            sql = (
                f"INSERT INTO `{_database}`.`{table}`({fields_str})\n"
                f"VALUES({fields_str_position})"
            )
        result = self.execute(sql, data, database, report=report)
        return result

==> reytool-0.66-py3-none-any.whl: reytool/rdatabase.py <==
# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
@Time    : 2022-12-05 14:10:02
@Author  : Rey
@Contact : reyxbo@163.com
@Explain : Rey's database connect type
"""


from typing import Any, List, Dict, Iterable, Optional, Literal, Union
import re
from sqlalchemy import create_engine, text
from sqlalchemy.engine.base import Engine, Connection
from sqlalchemy.engine.cursor import LegacyCursorResult
from sqlalchemy.sql.elements import TextClause

from .rbasic import get_first_notnull, error
from .rdata import to_table, to_df, to_json, to_sql, to_html, to_csv, to_excel
from . import roption
from .rtext import rprint
from .rwrap import runtime


def monkey_patch_more_fetch() -> None:
    """
    Add more methods to LegacyCursorResult object of sqlalchemy package.
    """

    # Fetch SQL result to table in List[Dict] format.
    LegacyCursorResult.fetch_table = to_table

    # Fetch SQL result to DataFrame object.
    LegacyCursorResult.fetch_df = to_df

    # Fetch SQL result to JSON string.
    LegacyCursorResult.fetch_json = to_json

    # Fetch SQL result to SQL string.
    LegacyCursorResult.fetch_sql = to_sql

    # Fetch SQL result to HTML string.
    LegacyCursorResult.fetch_sql = to_html

    # Fetch SQL result to save csv format file.
    LegacyCursorResult.fetch_csv = to_csv

    # Fetch SQL result to save excel file.
    LegacyCursorResult.fetch_excel = to_excel

monkey_patch_more_fetch()

class RConn(object):
    """
    Rey's database connect type.
    """

    # Values to be converted to None.
    none_values: List = ["", " ", b"", [], (), {}, set()]
    
    def __init__(
        self,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[str] = None,
        database: Optional[str] = None,
        charset: Optional[str] = None,
        once: bool = True,
        connect: Optional[Union[Engine, Connection]] = None
    ) -> None:
        """
        Set database connect parameters.
        """

        if type(connect) == Connection:
            connect = connect.engine
        if type(connect) == Engine:
            user = get_first_notnull(user, connect.url.username)
            password = get_first_notnull(password, connect.url.password)
            host = get_first_notnull(host, connect.url.host)
            port = get_first_notnull(port, connect.url.port)
            database = get_first_notnull(database, connect.url.database)
            charset = get_first_notnull(charset, connect.url.query.get("charset"))
            connect = connect.connect(close_with_result=once)
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.database = database
        self.charset = charset
        self.once = once
        self.connect = connect

    def create_conn(
        self,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[Union[str, int]] = None,
        database: Optional[str] = None,
        charset: Optional[str] = None,
        once: bool = None
    ) -> Connection:
        """
        Get database connect engine.
        """

        if self.connect != None\
            and (user == None or self.connect.engine.url.username == user) \
            and (password == None or self.connect.engine.url.password == password) \
            and (host == None or self.connect.engine.url.host == host) \
            and (port == None or self.connect.engine.url.port == port) \
            and (database == None or self.connect.engine.url.database == database) \
            and (charset == None or self.connect.engine.url.query["charset"] == charset):
            return self.connect

        user: str = get_first_notnull(user, self.user, default="error")
        password: str = get_first_notnull(password, self.password, default="error")
        host: str = get_first_notnull(host, self.host, default="error")
        port: Union[str, int] = get_first_notnull(port, self.port, default="error")
        database: str = get_first_notnull(database, self.database, default="error")
        charset: str = get_first_notnull(charset, self.charset, default="utf8")
        once: str = get_first_notnull(once, self.once, default=True)

        try:
            url = f"mysql+mysqldb://{user}:{password}@{host}:{port}/{database}?charset={charset}"
            engine = create_engine(url)
        except ModuleNotFoundError:
            url = f"mysql+pymysql://{user}:{password}@{host}:{port}/{database}?charset={charset}"
            engine = create_engine(url)
        connect = engine.connect(close_with_result=once)
        if self.connect == None and not once:
            self.connect = connect
        return connect
    
    def fill_sql_none(
        self,
        sql: Union[str, TextClause],
        params: Union[Dict, List[Dict]],
        fill_field: bool = True,
        none_values: List = none_values
    ) -> Union[Dict, List[Dict]]:
        """
        Fill missing parameters according to contents of sqlClause object of sqlalchemy module.

        Parameters
        ----------
        sql : SQL in sqlalchemy.text format or return of sqlalchemy.text.
        params : Parameters set for filling sqlalchemy.text.
        fill_field : Whether fill missing fields.
        none_values : Values to be converted to None.

        Returns
        -------
        Filled parameters.
            - When parameter params is Dict, then return Dict.
            - When parameter params is List[Dict], then return List[Dict].
        """

        if type(sql) == TextClause:
            sql = sql.text
        pattern = "(?<!\\\):(\w+)"
        sql_keys = re.findall(pattern, sql)
        if type(params) == dict:
            for key in sql_keys:
                if fill_field:
                    val = params.get(key)
                else:
                    val = params[key]
                if val in none_values:
                    val = None
                params[key] = val
        else:
            for param in params:
                for key in sql_keys:
                    if fill_field:
                        val = param.get(key)
                    else:
                        val = param[key]
                    if val in none_values:
                        val = None
                    param[key] = val
        return params

    def execute(
        self,
        sql: str,
        params: Optional[Union[List[Dict], Dict]] = None,
        database: Optional[str] = None,
        fill_field: bool = True,
        none_values: List = none_values,
        once: bool = None,
        report: bool = False,
        **kw_parms: Any
    ) -> LegacyCursorResult:
        """
        Execute SQL.

        Parameters
        ----------
        sql : SQL in sqlalchemy.text format.
        params : Parameters set for filling sqlalchemy.text.
        database : Database name.
        fill_field : Whether fill missing fields.
        none_values : Values to be converted to None.
        once : Whether the database connect engine is one-time (i.e. real time creation).
        report : Whether print SQL and SQL runtime.
        kw_parms : Keyword parameters for filling sqlalchemy.text.

        Returns
        -------
        LegacyCursorResult object of alsqlchemy package.
        """

        once = get_first_notnull(once, self.once, default=True)

        connect = self.create_conn(database=database, once=once)
        if params != None:
            if type(params) != list:
                params = [params]
            else:
                params = params.copy()
            for param in params:
                param.update(kw_parms)
        else:
            params = kw_parms
        if params != {}:
            params = self.fill_sql_none(sql, params, fill_field, none_values)
        sql = text(sql)
        if report:
            result, report_runtime = runtime(connect.execute, sql, params, _ret_report=True)
            report_info = "%s\nRow Count: %d" % (report_runtime, result.rowcount)
            if params == {}:
                rprint(report_info, sql, title="SQL", frame=roption.print_default_frame_half)
            else:
                rprint(report_info, sql, params, title="SQL", frame=roption.print_default_frame_half)
        else:
            result = connect.execute(sql, params)
        if once:
            connect.close()
            del self.connect
            self.connect = None
        return result

    def execute_select(
            self,
            table: str,
            database: Optional[str] = None,
            fields: Optional[Union[str, Iterable]] = None,
            where: Optional[str] = None,
            order: Optional[str] = None,
            limit: Optional[Union[int, str, Iterable[Union[int, str]]]] = None,
            report: bool = False
        ) -> LegacyCursorResult:
        """
        Execute select SQL.

        Parameters
        ----------
        table : Table name.
        database : Database name.
        fields : Select syntax content.
            - None : Is 'SELECT *'.
            - str : Join as 'SELECT str'.
            - Iterable[str] : Join as 'SELECT \`str\`, ...'.

        where : 'WHERE' syntax content, join as 'WHERE str'.
        order : 'ORDER BY' syntax content, join as 'ORDER BY str'.
        limit : 'LIMIT' syntax content.
            - Union[int, str] : Join as 'LIMIT int/str'.
            - Iterable[Union[str, int]] with length of 1 or 2 : Join as 'LIMIT int/str, ...'.

        report : Whether print SQL and SQL runtime.

        Returns
        -------
        LegacyCursorResult object of alsqlchemy package.
        """

        sqls = []
        if database == None:
            _database = self.database
        else:
            _database = database
        if fields == None:
            fields = "*"
        elif type(fields) != str:
            fields = ",".join(["`%s`" % field for field in fields])
        select_sql = (
            f"SELECT {fields}\n"
            f"FROM `{_database}`.`{table}`"
        )
        sqls.append(select_sql)
        if where != None:
            where_sql = "WHERE %s" % where
            sqls.append(where_sql)
        if order != None:
            order_sql = "ORDER BY %s" % order
            sqls.append(order_sql)
        if limit != None:
            list_type = type(limit)
            if list_type in [str, int]:
                limit_sql = f"LIMIT {limit}"
            else:
                if len(limit) in [1, 2]:
                    limit_content = ",".join([str(val) for val in limit])
                    limit_sql = "LIMIT %s" % limit_content
                else:
                    error("The length of the limit parameter value must be 1 or 2", ValueError)
            sqls.append(limit_sql)
        sql = "\n".join(sqls)
        result = self.execute(sql, database=database, report=report)
        return result

    def execute_update(
        self,
        data: Union[LegacyCursorResult, List[Dict], Dict],
        table: str,
        database: Optional[str] = None,
        primary_key: Optional[str] = None,
        report: bool = False
    ) -> LegacyCursorResult:
        """
        Update the data of table in the datebase.
    
        Parameters
        ----------
        data : Updated data.
        table : Table name.
        database : Database name.
        primary_key : 'WHERE' syntax content.
            - None : Is 'WHERE `%s` = :%s' % (first field name and value of each item).
            - str : Is f'WHERE `{str}` = :{str}'.

        report : Whether print SQL and SQL runtime.

        Returns
        -------
        LegacyCursorResult object of alsqlchemy package.
        """

        data_type = type(data)
        if data_type == LegacyCursorResult:
            data = to_table(data)
        elif data_type == dict:
            data = [data]
        if database == None:
            _database = self.database
        else:
            _database = database
        data_flatten = {}
        sqls = []
        for index, row in enumerate(data):
            for key, val in row.items():
                index_key = "%d_%s" % (index, key)
                data_flatten[index_key] = val
            if primary_key == None:
                primary_key = list(row.keys())[0]
            set_content = ",".join(
                [
                    "`%s` = :%d_%s" % (key, index, key)
                    for key in row
                    if key != primary_key
                ]
            )
            sql = (
                f"UPDATE `{_database}`.`{table}`\n"
                f"SET {set_content}\n"
                f"WHERE `{primary_key}` = :{index}_{primary_key}"
            )
            sqls.append(sql)
        sqls = ";\n".join(sqls)
        result = self.execute(sqls, data_flatten, database, once=False, report=report)
        if self.once:
            self.connect == None
        return result

    def execute_insert(
        self,
        data: Union[LegacyCursorResult, List[Dict], Dict],
        table: str,
        database: Optional[str] = None,
        duplicate_method: Optional[Literal["ignore", "update"]] = None,
        report: bool = False
    ) -> LegacyCursorResult:
        """
        Insert the data of table in the datebase.

        Parameters
        ----------
        data : Updated data.
        table : Table name.
        database : Database name.
        duplicate_method : Syntax method when constraint error.
            - None : Then no syntax.
            - 'ignore' : Is 'UPDATE IGNORE INTO'.
            - 'update' : Is 'ON DUPLICATE KEY UPDATE'.

        report : Whether print SQL and SQL runtime.

        Returns
        -------
        LegacyCursorResult object of alsqlchemy package.
        """

        data_type = type(data)
        if data_type == LegacyCursorResult:
            data = self.to_table(data)
        elif data_type == dict:
            data = [data]
        if database == None:
            _database = self.database
        else:
            _database = database
        fields = list({key for row in data for key in row})
        fields_str = ",".join(["`%s`" % field for field in fields])
        fields_str_position = ",".join([":" + field for field in fields])
        if duplicate_method == "ignore":
            sql = (
                f"INSERT IGNORE INTO `{_database}`.`{table}`({fields_str})\n"
                f"VALUES({fields_str_position})"
            )
        elif duplicate_method == "update":
            update_content = ",".join(["`%s` = VALUES(`%s`)" % (field, field) for field in fields])
            sql = (
                f"INSERT INTO `{_database}`.`{table}`({fields_str})\n"
                f"VALUES({fields_str_position})\n"
                "ON DUPLICATE KEY UPDATE\n"
                f"{update_content}"
            )
        else:
            sql = (
                f"INSERT INTO `{_database}`.`{table}`({fields_str})\n"
                f"VALUES({fields_str_position})"
            )
        result = self.execute(sql, data, database, report=report)
        return result
